;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                       ;;;
;;;  define-record and record-case macros for guile                       ;;;
;;;                                                                       ;;;
;;;                                                                       ;;;
;;;  Implemented by David Lutterkort (lutterdc@cs.purdue.edu) borrowing   ;;;
;;;  heavily from previous work by Jens Palsberg, Duncan Fisher,          ;;; 
;;;  Brent Benson, Gary Duzan, David McCusker, and Dan Ryan.              ;;;
;;;                                                                       ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-module (simplerec)
  :use-module (ice-9 common-list))

(define (format-> x l)
  (string->symbol (string-append (symbol->string x) "->" (symbol->string l))))

(define (format? x)
  (string->symbol (string-append (symbol->string x) "?")))

(define (formatmake- x)
  (string->symbol (string-append "make-" (symbol->string x))))

(define (format-rtd name)
  (string->symbol (string-append (symbol->string name) "-rtd")))

(define (format! name field)
  (string->symbol
   (string-append (symbol->string name) "-set-" (symbol->string field))))

;;;
;;; define-record macro
;;;
;;; Syntax of general form: (define-record name (field-1 ... field-n))
;;;
;;; This macro expansion defines the following procedures:
;;;    make-name
;;;    name->field-1
;;;    ...
;;;    name->field-n
;;;    name-set-field-1
;;;    ...
;;;    name-set-field-n
;;;    name?
;;;

(defmacro-public define-record args   ;; args = (name . fields)
(let ((name (car args))
      (fields (cadr args)))
  (let* ((type-name   (format-rtd name))
	 (predicate   (format? name))
	 (constructor (formatmake- name))
	 (make-reader
	  (lambda (field)
	    `(define ,(format-> name field) 
	       (record-accessor ,type-name ',field))))
	 (make-writer
	  (lambda (field)
	    `(define ,(format! name field)
	       (record-modifier ,type-name ',field)))))
    `(begin
       (define ,type-name (make-record-type ',name ',fields))
       (define ,predicate (record-predicate ,type-name))
       (define ,constructor (record-constructor ,type-name))
       ,@(map make-reader fields)
       ,@(map make-writer fields)))))

;;;
;;; record-case macro
;;;
;;; Syntax of general form: (record-case record-expression
;;;                           (name-1 field-list-1 consequent-1)
;;;                           ...
;;;                           (name-n field-list-n consequent-n)
;;;                           (else alternative))
;;;
;;; For each clause in the record-case statement, the record-case
;;; macro expands it to a cond type clause of the form ((pred? value)
;;; (let-expression)), where the body of the let expression is the
;;; consequent expressions of the record-case form. The bindings of the
;;; let form are those between the symbols of the record-case field-list
;;; and the corresponding values of the record passed in as argument
;;; record-var to record-case. These bindings are done by field-name, 
;;; i.e. field-list-i doesn't need to contain all fields.
;;;

(defmacro-public record-case args    ;; args = (record-exp . clauses)
(let ((record-exp (car args))
      (clauses (cdr args)))
  (let ((var     (gensym)))
    ;; make-conds expands all the record-case clauses into a
    ;; list of appropriate cond style clauses, i.e. each clause
    ;; list of the form:
    ;;    (name-i field-list-i consequent-i)
    ;; gets expanded into:
    ;;    ((name-i? record-exp)
    ;;        (let ((field-list-i-1 field-name-1)
    ;;              (           ...             )
    ;;              (field-list-i-n field-name-n))
    ;;          consequent-i))
    (letrec ((make-cond
             (lambda (clause)
               (cond
                ((eq? (car clause) 'else)
                 `((else ,@(cdr clause))))
                (else
    ;; The make-lets procedure returns the let expression bindings for
    ;; one record-case clause only. These in turn will be combined with
    ;; the (pred? value) list to form one clause of the cond statement.
		 (let* ((key (car clause))
			(pred? `(,(format? key) ,var))
			(make-let
			 (lambda (field)
			   (let ((get (format-> key field)))
			     `(,field (,get ,var))))) 
			(lets (map make-let (cadr clause))))
		   `(,pred? (let ,lets ,@(cddr clause)))))))))
      (if (record-macros:check-syntax clauses)
	  `(let ((,var ,record-exp)) (cond ,@(map make-cond clauses)))
	  ;; this error call should never be processed since
	  ;; check-syntax only returns #t or doesn't return at all
	  (error 'record-case-macro
		 "Unspecified error ~%Record-exp = ~p~%Clauses = ~p"
		 record-exp clauses))))))

;;;
;;; check-syntax analyses the syntax of the clauses of the record-case
;;; statement, e.g. is it a list of lists, are the cars of each list a
;;; symbol or set of symbols, are the second elements of each list also
;;; a set of symbols, etc. Also checks to see if the keys to all the
;;; clauses are unique. Returns #t if there are no syntax errors found,
;;; otherwise it calls the error procedure and doesn't return.
;;;

(define (record-macros:check-syntax clauses)
  ;; clause? returns true if a single record-case
  ;; clause has the correct syntax, otherwise false
  (let ((clause? (lambda (c)
                   (and (list? c)
                        (not (null? c))
                        (or (symbol? (car c))); (has-duplicates? (car c)))
                        (if (eq? 'else (car c))
                          (not (null? (cdr c)))
                          (and (> (length c) 2)
                               (not (has-duplicates? (cadr c)))))))))
  ;; check-clauses returns #t if there are no invalid clauses detected,
  ;; otherwise the error procedure is invoked and no value is returned
    (letrec ((check-clauses
              (lambda (clauses)
                (cond
                 ((null? clauses) #t)
                 ((not (clause? (car clauses)))
                  (error 'record-macros:check-syntax
                    "~%Incorrect syntax, clause invalid:~%~p" (car clauses)))
                 ((and (eq? (caar clauses) 'else)
                       (not (null? (cdr clauses))))
                  (error 'record-macros:check-syntax
                    "~%Incorrect syntax, clauses after else:~%~p" clauses))
                 (else (check-clauses (cdr clauses))))))
  ;; case-names returns a list containing all the key symbols
  ;; of all the clauses of the record-case (including else)
             (case-names
              (lambda (clauses)
                (cond
                 ((null? clauses) '())
                 ((list? (caar clauses))
                  (append (caar clauses) (case-names (cdr clauses))))
                 (else
                  (cons (caar clauses) (case-names (cdr clauses))))))))
      (let ((keys (case-names clauses)))
        (if (not (has-duplicates? keys))
          (check-clauses clauses)
          (error 'record-macros:check-syntax
            "~%Incorrect syntax, duplicate clause keys in~%~p" keys))))))


;;;; Test
;(define-record r1 (a b))
;(define-record r2 (x y z))
;(define s1 (make-r1 'a1 'b1))
;(define s2 (make-r2 'x1 'y1 'z1))

;(record-case
; s2
; (r1 (a b) (pk 'r1 b a))
; (r2 (x z) (pk 'r2 x z)))
