;;; wickel.scm -- wrap command line tools with a GUI
;;;
;;; Copyright (C) 1998 David Lutterkort <lutter@cs.purdue.edu>
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2, or (at
;;; your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with this software; see the file COPYING.  If not, write to
;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,
;;; USA.
;;;
;;;
;;; Annotated grammar for a wickel file
;;; ===================================
;;; [Note: I use EBNF, i.e. '::=', '[' ,']' ,'|' are meta-symbols
;;;        Terminals are all-lower case whereas non-terminals are uppercase ]
;;;
;;; CLIWRAP ::= PROGRAM DECL+ LAYOUT
;;;     A wickel file consists of two parts:
;;;     (1) The declarations (DECL)
;;;     (2) A description of the layout
;;;
;;; PROGRAM ::= (program NAME PROPS)
;;;   The first declaration in a wickel file has to be a PROGRAM statement.
;;;   The PROPS in it is a list describing some special things about the 
;;;   program:
;;;   (format [cram-switches] [separate-bool])
;;;     the format clause regulates how options are output:
;;;     o if 'cram-switches is present, there will be no space between
;;;       a switch and its argument (-Pprinter), otherwise there 
;;;       will (-P printer)
;;;     o if 'separate-bool is present, on/off switches won't be combined 
;;;       to one big switch (ls -l -g -t -m vs. ls -lgtm)
;;;   (path STRING)
;;;     Specifies the path from which the described program comes, e.g.
;;;     (path "/usr/local/X11/bin") If no path is given, the command is 
;;;     searched on the standard path.
;;;
;;; DECL    ::= SWITCH | ENUM
;;; ENUM    ::= (enum NAME TYPE STRINGS+)
;;;   Introduces a new type called NAME, whose elements are
;;;   of the basetype TYPE and have the values STRINGS+ (at the moment, 
;;;   only enumerations consisting ofg  strings can be created.
;;;
;;; SWITCH ::= (switch NAME OPTION [TYPE] LABEL [HELP])
;;;          | (environ NAME [TYPE] [LABEL [HELP]])
;;;          | (group NAME NAME SWITCH_REF+ LABEL)
;;;   o switch creates a command line switch with name NAME, corresponding to 
;;;     the option OPTION of type TYPE (if it is not given, bool is assumed)
;;;     The input field in the GUI wil have the label LABEL and get the 
;;;     helptext HELP
;;;   o environ declares an environment variable called NAME.
;;;   o group declares a group of switches that can be referred to with the
;;;     first NAME, the second NAME is the default and SWITCH_REF+ is a list
;;;     of switches that are mutually exclusive
;;;     
;;;  
;;; SWITCH_REF ::= NAME | SWITCH
;;; OPTION ::= NAME
;;;   This is not quite true: any letter is a valid option. 
;;;   #f is also allowed to indicate a no-op (i.e., no switch will be output
;;;   by the corresponding field.
;;;
;;; TYPE ::= TYPE_NAME | TYPE_NAME SIZE
;;; TYPE_NAME ::= NAME | string | integer | bool | file
;;; SIZE ::= NUMBER | #f | #t
;;; HELP ::= STRING
;;;
;;; LAYOUT ::= (layout PAGE+)
;;; PAGE ::= (page STRING ELEMENT)
;;; ELEMENT ::= (hbox [TITLE] ELEMENT+) | (vbox [TITLE] ELEMENT+) | NAME
;;; 
;;; NAME is any valid scheme symbol
;;; STRING is a valid scheme string

(use-modules (simplerec) (ice-9 common-list))
(load-from-path "gtk-utils.scm")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Records representing the abstract syntax tree
;;;
(define-record type-asn (type size))
(define-record switch-asn (option type label help))
;; SWITCH-ASN is used to describe both, command line switches and
;; environment variables. If the switch in question is really an 
;; environment variable, OPTION is '*environ*
(define (make-environ-asn type label help)
  (make-switch-asn '*environ* type label help))
(define (switch-is-environ? asn)
  (and (switch-asn? asn) (eq? (switch-asn->option asn) '*environ*)))

(define-record group-asn (default switch+ label))
(define-record program (name args path format))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Functions and macros for parsing the declaration part of
;;; an input file
;;;
(define (parse-type type)
  ;; type ::= #f | (BUILTIN_TYPE) | (BUILTIN_TYPE SIZE)
  (let* ((type-name (if type (car type) 'bool))
	 (type-rec (type-ref type-name))
	 (size (if (and type (not (null? (cdr type))))
		   (cadr type)
		   (type->size type-rec))))
    (make-type-asn type-rec size)))
      
(define (parse-help args)
  (and (not (null? args)) (car args)))

(define (parse-option opt)
  (cond
   ((symbol? opt) (symbol->string opt))
   ((integer? opt) (number->string opt))
   ((or (string? opt) (eq? opt #f)) opt)
   (else
    (error 'invalid-option-specifier opt))))

(define (parse-type-label-help name args)
  ;; parse the [TYPE] LABEL HELP part for environ and switch declarations
  (let loop ((args args) (type #f))	
    (cond
     ((null? args) 
      (list (parse-type type) (symbol->string name) (parse-help '())))
     ((string? (car args))
      ;; Found the label
      (list (parse-type type) (car args) (parse-help (cdr args))))
     ((not type)
      (loop (cdr args) (list (car args))))
     (else
      (loop (cdr args) (append type (list (car args))))))))

(define (parse-switch name opt args)
  (apply make-switch-asn (cons (parse-option opt)
			       (parse-type-label-help name args))))


(define (parse-environ name args)
  (apply make-environ-asn (parse-type-label-help name args)))

;;;
;;; The macros that are used in the user-supllied GUI description
;;; This is the only 'public' part of this file
;;;

(defmacro program (proggy . desc)
  `(let ((format (assq-ref ',desc 'format))
	 (path (assq-ref ',desc 'path)))
     (set! symtab '())
     (set! fieldtab '())
     (set! typetab typetab-init)
     (symtab-add '*program*
		 (make-program
		  (if (string? ',proggy) ',proggy (symbol->string ',proggy))
		  'file
		  (if path (car path) "") 
		  (or format '())))))

(defmacro switch (name opt . rest)
  ;; The start of rest might be the type
  `(symtab-add ',name (parse-switch ',name ',opt ',rest)))

(defmacro environ (name . rest)
  `(symtab-add ',name (parse-environ ',name ',rest)))

(defmacro enum (name type . values)
  `(cond
    ((not (eq? ',type 'string))
     (error "Basetype for enum has to be 'string" ',type))
    ((notevery string? ',values)
     (error "The enumeration values have to be strings" ',values))
    (else
     (type-add-enum ',name ',type ',values))))

(defmacro group (name default switch+ label)
  ;; Process the switches in switch+ and collect their names 
  ;; Should this be a layout (??)
  (let ((get-name 
	 (lambda (switch)
	   (if (symbol? switch) switch (eval switch))))
	(not-is-switch
	 (lambda (name)
	   (not (switch-asn? (symtab-ref name))))))
  `(let ((names (map ,get-name ',switch+)))
     (if (or (not ',default) (memq ',default names))
	 (let ((bad (pick ,not-is-switch names)))
	   (if (null? bad)
	       (symtab-add ',name 
			   (make-group-asn ',default names ,label))
	       (error "Group can only contain switches" ',name bad)))
	 (error "Default value must be in group"  ',name ',default)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Macros for processing the layout
;;;

(defmacro layout page+
  ;; Here's where the meet should happen
  `(parse-layout ',page+))

(define tooltips #f)
(define (parse-layout pages)
  (set! tooltips (gtk-tooltips-new))
  (create-layout (map parse-page pages)))

(define (parse-page a-page)
  (if (and (list? a-page) (eq? (car a-page) 'page)
	   (string? (cadr a-page)))
      (create-page (cadr a-page) (parse-element (caddr a-page)))
      (error 'not-a-page a-page)))

(define (parse-element elt)
  (cond
   ((symbol? elt) (create-field elt))
   ((and (list? elt) (memq (car elt) '(hbox vbox)))
    (parse-box (car elt) (cdr elt)))
   (else (error 'not-an-element elt))))

(define (parse-box kind elts)
  (let* ((title (if (string? (car elts)) (car elts) #f))
	 (elt-list (if title (cdr elts) elts)))
    (create-box kind title
		(map parse-element elt-list))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Creating the widgets and hooking them up
;;;
(define (create-layout pages)
  ;; pages is a list of  (title . widget) pairs
  (let ((window (gtk-window-new 'toplevel))
	(box (gtk-vbox-new #f 0))
	(button-box (gtk-hbox-new #t 0))
	(quit-button (gtk-button-new-with-label "Quit"))
	(run-button (gtk-button-new-with-label "Execute"))
	(clip-button (gtk-button-new-with-label "Clipboard"))
	(notebook (gtk-notebook-new))
	(args (make-file-field "Arguments:" 20 #f)))

    (define (quit-thunk) (gtk-widget-destroy window))
    (define (read-values)
      (define (form-command bool other)
	(let ((bool (if (string=? bool "") "" (string-append " -" bool)))
	      (progname (program->name (symtab-ref '*program*)))
	      (path  (program->path (symtab-ref '*program*)))
	      (args (or ((field->reader args)) "")))
	  (let ((cmd
		 (string-append path progname bool other " " args)))
	    (display cmd)
	    (newline)
	    cmd)))
      (let* ((format (program->format (symtab-ref '*program*)))
	     (s (if (memq 'cram-switches format) "" " "))
	     (c (memq 'separate-bool format)))
	(let loop ((fields fieldtab)
		   (bool "")
		   (other " "))
	  (if (null? fields)
	      (form-command bool other)
	      (let* ((name (caar fields))
		     (field (cdar fields))
		     (asn (symtab-ref name)))
		(if (switch-asn? asn)
		    (let ((char (switch-asn->option asn))
			  (val ((field->reader field))))
		      (cond 
		       ((or (not char) (not val))
			(loop (cdr fields) bool other))
		       ((and (string=? "" val) (not c))
			(loop (cdr fields) (string-append bool char)
			      other))
		       (else
			(loop (cdr fields) bool
			      (string-append other " -" char s val)))))
		    (loop (cdr fields) bool other)))))))
    (define (execute)
      ;; Very, very crude
      ; (system (read-values))
      (display "exec ")
      (read-values))

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; Set up the GUI
    (for-each
     (lambda (x)
       (let ((label (gtk-label-new (car x))))
	 (gtk-widget-show label)
	 (gtk-notebook-append-page notebook (cdr x) label)))
     pages)
	 
    ;; Attach callbacks to the various elements
    (gtk-signal-connect quit-button "clicked" quit-thunk)
    (gtk-signal-connect clip-button "clicked" read-values)
    (gtk-signal-connect run-button "clicked" execute)
    (gtk-signal-connect window "delete_event" (lambda (ev) #t))
    (gtk-signal-connect window "destroy" gtk-main-quit)

    ;; Make the button row at the bottom
    (gtk-box-pack-start button-box run-button #t #t 0)
    (gtk-box-pack-start button-box clip-button #t #t 0)
    (gtk-box-pack-start button-box quit-button #t #t 0)
    (gtku-widget-show* button-box run-button clip-button quit-button)
    ;; Pack the notebook and button-box into one vbox
    (gtk-box-pack-start box notebook #t #t 0)
    (gtk-box-pack-start box (field->widget args) #f #f 5)
    (gtk-box-pack-start box button-box #f #f 0)
    (gtk-container-add window box)
    (gtk-window-set-title 
     window (string-append "Wickel: " (program->name (symtab-ref '*program*))))
    (gtku-widget-show* notebook box window)
    (gtk-main)))
    
(define create-page cons)
  
(define (create-box kind title widgets)
  (let ((box (case kind
	       ((hbox) (gtk-hbox-new #f 0))
	       ((vbox) (gtk-vbox-new #f 0)))))
    (for-each (lambda (elt) (gtk-box-pack-start box elt #f #f 0)) widgets)
    (gtk-widget-show box)
    (if title
	(let ((frame (gtk-frame-new title)))
	  (gtk-widget-show frame)
	  (gtk-container-add frame box)
	  frame)
	box)))
    
(define (create-field name)
  (let ((record (symtab-ref name)))
    (record-case 
     record
     (switch-asn () (create-switch-field name record 'none))
     (group-asn () (create-group-field name record)))))

(define (create-switch-field name record group)
  (let* ((type (switch-asn->type record))
	(label (switch-asn->label record))
	(widget
	 (add-field (type-asn->type type) name label 
		    (type-asn->size type) group)))
    (if (switch-is-environ? record)
	(let ((default (getenv name))
	      (field (fieldtab-ref name)))
	  (field-set-reader field (lambda () #f))
	  (if default  ((field->setter field) default))))
    widget))

(define (create-group-field name record)
  (let ((default (group-asn->default record))
	(box (gtk-vbox-new #f 0)))
    (let loop ((switch+ (group-asn->switch+ record)) (group #f))
      (if (not (null? switch+))
	  (let* ((wname (car switch+))
		 (switch (symtab-ref wname))
		 (widget (create-switch-field wname switch group))
		 (radio (add-field-to-group wname group)))
	    (if (eq? default wname)
		(gtk-toggle-button-set-state radio #t))
	    (gtk-box-pack-start box 
				(field->widget (fieldtab-ref wname)) #f #f 0)
	    (loop (cdr switch+) radio))))
    (gtk-widget-show box)
    (if (group-asn->label record)
	(let ((frame (gtk-frame-new (group-asn->label record))))
	  (gtk-widget-show frame)
	  (gtk-container-add frame box)
	  frame)
	box)))
    
;; Field table:
;;   Contains the abstract description of every input field, indexed by NAME
;;   The desc is : the widget, a reader function and a setter function
;;   to read the value in an input field and to set it to a new value
(define-record field (widget reader setter))
(define fieldtab '())
(define (fieldtab-add name field)
  (set! fieldtab (assq-set! fieldtab name field)))
(define (fieldtab-ref name)  (assq-ref fieldtab name))

(define (add-field type name label width radio)
  (case (type->tag type)
    ((bool) (add-bool-field name label radio))
    ((string) (add-string-field name label width))
    ((integer) (add-integer-field name label width))
    ((file) (add-file-field name label width))
    ((enum) (add-enum-field name label width))
    (else (error "Unknown type in add-field:" type))))


(define (add-bool-field name label radio)
  (let* ((button 
	  (if (eq? radio 'none)
	      (gtk-check-button-new-with-label label) 
	      (gtk-radio-button-new-with-label radio label)))
	 (var #f)
	 (help (symtab->help name))
	 (reader (lambda () (if (gtk-toggle-button-active button) "" #f)))
	 (setter (lambda (val) (gtk-toggle-button-set-state button val))))
    (gtk-widget-show button)
    (if help
	(gtk-tooltips-set-tip tooltips button help "??bool??"))
    (fieldtab-add name (make-field button reader setter))
    button))

(define (add-field-to-group name group)
  (if (eq? (type->tag (symtab->type name)) 'bool)
      (field->widget (fieldtab-ref name))
      (let* ((field (fieldtab-ref name))
	     (box (field->widget field))
	     (newbox (gtk-hbox-new #f 0))
	     (val-reader (field->reader field))
	     (button (gtk-radio-button-new group))
	     (callback
	      (lambda ()
		(let ((val (gtk-toggle-button-active button)))
		  (gtk-widget-set-sensitive box val))))
	     (reader 
	      (lambda ()
		(if (gtk-toggle-button-active button)
		    (val-reader)
		    #f))))
	  (gtk-signal-connect button "clicked" callback)
	  (callback)
	  (gtk-box-pack-start newbox button #f #f 0)
	  (gtk-box-pack-start newbox box #t #t 0)
	  (field-set-reader field reader)
	  (field-set-widget field newbox)
	  (gtku-widget-show* button newbox)
	  button)))

(define (add-string-field name text width)
  ;; Generate label + entry in an hbox
  (let* ((label (gtk-label-new text))
	 (entry (gtk-entry-new))
	 (help (symtab->help name))
	 (box (gtk-hbox-new #f 0))
	 (reader 
	  (lambda () 
		   (let ((val (gtk-entry-get-text entry)))
		     (if (not (string=? "" val))
			 (string-append "\"" (gtk-entry-get-text entry) "\"")
			 #f))))
	 (setter (lambda (val) (gtk-entry-set-text entry val))))
    
    (if help
	(gtk-tooltips-set-tip tooltips entry help "??entry??"))
    (gtk-box-pack-start box label #f #f 5)
    (gtk-box-pack-start box entry #t (not (not width)) 0)
    (gtku-widget-show* box label entry)
    (fieldtab-add name (make-field box reader setter))
    box))

(define (add-integer-field name text width)
  ;; Generate label + entry in an hbox
  (let* ((label (gtk-label-new text))
	 (entry (gtk-entry-new))
	 (help (symtab->help name))
	 (box (gtk-hbox-new #f 0))
	 (validator 
	  (lambda ()
	    (if (not (integer? (string->number (gtk-entry-get-text entry))))
		(pk 'not-valid text))))
	 (reader (lambda () 
		   (let ((val (gtk-entry-get-text entry)))
		   (if (integer? (string->number val))
		       val
		       #f))))
	 (setter (lambda (val)
		   (gtk-entry-set-text entry (number->string val)))))
    (gtk-box-pack-start box label #f #f 5)
    (gtk-box-pack-start box entry #f #f 0)
    (if help
	(gtk-tooltips-set-tip tooltips box help "??box??"))
    (gtku-widget-show* box label entry)
    (gtk-signal-connect entry "activate" validator)
    (fieldtab-add name (make-field box reader setter))
    box))

(define (add-enum-field name text width)
  (let* ((enum (type->specs (symtab->type name)))
	 (values (type-enum-specs->values enum))
	 (label (gtk-label-new text))
	 (combo (gtk-combo-new))
	 (entry (gtk-combo-entry combo))
	 (help (symtab->help name))
	 (box (gtk-hbox-new #f 0))
	 (reader 
	  (lambda () 
		   (let ((val (gtk-entry-get-text entry)))
		     (if (not (string=? "" val))
			 (string-append "\"" (gtk-entry-get-text entry) "\"")
			 #f))))
	 (setter (lambda (val) (gtk-entry-set-text entry val))))
    
    (if help
	(gtk-tooltips-set-tip tooltips entry help "??entry??"))
    (gtk-box-pack-start box label #f #f 5)
    (gtk-box-pack-start box combo #t (not (not width)) 0)
    (gtk-combo-set-popdown-strings combo values)
    (gtk-entry-set-text entry "")
    (gtku-widget-show* box label combo)
    (fieldtab-add name (make-field box reader setter))
    box))

(define (make-file-field text width help)
  (let* ((label (gtk-label-new text))
	 (entry (gtk-entry-new))
	 (browse (gtk-button-new-with-label "Browse..."))
	 (box (gtk-hbox-new #f 0))
	 (reader 
	  (lambda () 
	    (gtk-entry-get-text entry)))
	 (setter (lambda (val) (gtk-entry-set-text entry val)))
	 (browse-cb 
	  (lambda ()
	    (let ((fname (gtku-file-select 
			  (string-append "Browse for " text) "")))
	      (if fname
		  (gtk-entry-set-text entry fname))))))
    
    (if help
	(gtk-tooltips-set-tip tooltips entry help "??entry??"))
    (gtk-box-pack-start box label #f #f 5)
    (gtk-box-pack-start box entry #t (not (not width)) 0)
    (gtk-box-pack-start box browse #f #f 3)
    (gtk-signal-connect browse "clicked" browse-cb)
    (gtku-widget-show* label entry browse box)
    (make-field box reader setter)))
    
(define (add-file-field name text width)
  ;; Generate label + entry in an hbox
  (let ((field (make-file-field text width (symtab->help name))))
    (fieldtab-add name field)
    (field->widget field)))
  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  
;; Symbol table:
;;   only one needed -- all names are global
;;   Implement as simple alist (names have to be scheme symbols)
(define symtab '())
(define (symtab-add name entry)
  (cond
   ((assq name symtab)
    (error "Symbol already defined:" name))
   (else
    (set! symtab (acons name entry symtab))
    name)))
(define (symtab-find name)  
  (or (assq name symtab)
      (error "Refernce to unknown symbol:" name)))
(define (symtab-ref name) 
  (or (assq-ref symtab name)
      (error "Reference to unknown symbol:" name)))
(define (symtab->type-asn name)
  (let ((asn (symtab-ref name)))
    (if asn
	(switch-asn->type asn)
	#f)))

(define (symtab->type name)
  ;; Extract the type of an entry
  (let ((type-asn (symtab->type-asn name)))
    (if type-asn (type-asn->type type-asn) #f)))

(define (symtab->size name)
  ;; Extract the size of an entry
  (let ((type-asn (symtab->type-asn name)))
    (if type-asn (type-asn->size type-asn) #f)))

(define (symtab->help name)
  ;; Extract the help string
  (let ((asn (symtab-ref name)))
    (if asn
	(switch-asn->help asn)
	#f)))
    
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Entries in the type table
(define-record type (tag size specs))
;; TAG is a symbol. For builtin types, specs is #f
;; for other types it contains additional information
(define-record type-enum-specs (basetype values))

(define typetab-init
  `((bool    . ,(make-type 'bool #f #f))
    (string  . ,(make-type 'string 20 #f))
    (integer . ,(make-type 'integer 6 #f))
    (file    . ,(make-type 'file 20 #f))))

(define typetab typetab-init)

(define (type-add name type)
  (set! typetab (acons name type typetab)))

(define (type-add-enum name basetype values)
  (type-add name (make-type 'enum 20 (make-type-enum-specs basetype values))))

(define (type-find name)
  (or (assq name typetab)
      (error "Reference to unknown type:" name)))

(define (type-ref name)
  (or (assq-ref typetab name)
      (error "Reference to unknown type:" name)))
